#!/usr/freeware/bin/wish -f
########################################################################
#                                                                      #
# update                  Tcl/Tk Script               File: update.tcl #
#                                                                      #
# Tcl/Tk script to update selected parts or subparts of the Breitenau  #
# grade octree                                                         #
#                                                                      #
# Date written  :  3/05/2000 by R. Prissang (FUB)                      #
# Last modified : 17/02/2022 (15:30) by R. Prissang (FUB)              #
#                                                                      #
########################################################################

# Request required packages
package require msgcat
package require Pmstools

# Define project code
set Prj "BR"

# Add the name of the PMS tool library to the autoload path
lappend auto_path [file join $env(OPMOD_HOME_$Prj) lib scripts]

# Include the submodule of script update
source [file join $env(OPMOD_HOME_$Prj) update1.tcl]
source [file join $env(OPMOD_HOME_$Prj) update2.tcl]
source [file join $env(OPMOD_HOME_$Prj) update3.tcl]
source [file join $env(OPMOD_HOME_$Prj) update4.tcl]
source [file join $env(OPMOD_HOME_$Prj) update5.tcl]

# Adjust locale to a user defined setting
adjustLocale OPMOD_LANG_$Prj

# Determine name of current script
set ScriptName [file rootname [file tail [info script]]]

# Load message catalog for the application
loadMsgCatalog [file join $env(OPMOD_HOME_$Prj) msgs $ScriptName]

# Load message catalog for package Pmstools
loadPmsMsgCatalog

# Set project's configuration directory
if [info exists env(OPMOD_PROJ_$Prj)] {
  set PrjConfigDir [file join $env(OPMOD_PROJ_$Prj) "config"]
} else {
  set PrjConfigDir "config"
}

# Read project information
getProjectInfo $PrjConfigDir PrjInfo
getUniverseInfo $PrjConfigDir PrjInfo
getAttrInfo $PrjConfigDir ModelAttrList
getRegionInfo $PrjConfigDir RegionList
array set PrjInfo { \
 DataFileCols \
 { {bid T "booking number" 10 {-}} \
   {x N "x coordinate" 13 {}} \
   {y N "y coordinate" 11 {}} \
   {z N "z coordinate" 9 {}} \
   {thickn N "thickness" 6 {}} \
   {to N "tonnage" 7 {-1}} \
   {ca N "CaO grade" 6 {-1}} \
   {si N "SiO2 grade" 6 {-1}} \
   {fe N "Fe2O3 grade" 6 {-1}} \
   {spw N "spatter value" 6 {-1}} \
   {fault N "fault indicator" 6 {-1}} \
   {bau T "bau" 9 {?}} \
   {fenster T "fenster" 7 {?}} \
   {date G "date" 10 {- ?}} \
 } \
}

# Set project name
set ProjectName [::msgcat::mc $PrjInfo(ProjectName)]

# Set control parameters for the main window
wm title . [::msgcat::mc "Update Monitor (%s)" $ProjectName]
wm iconname . [::msgcat::mc "Monitor"]

# Execute the exit command in case the root window is closed by the user
wm protocol . WM_DELETE_WINDOW {exit}

# Manage input focus to comply with the window manager's style guide
manageTkFocus . force

# Get resources from a configuration file on a non-Unix system
getResources [file join $env(OPMOD_HOME_$Prj) config Xresources]

# Get title and monitor fonts from the option database
set TitleFont [option get . titleFont Font]
set MonitorFont [option get . monitorFont Font]

# Change working directory to the Operative Modelling project directory
if {[info exists env(OPMOD_PROJ_$Prj)]} {
  cd $env(OPMOD_PROJ_$Prj)
}

# Create menu bar
. configure -menu .mBar
menu .mBar

# Create file menu
.mBar add cascade -label [::msgcat::mc "File"] -underline 0 \
  -menu .mBar.file
menu .mBar.file -tearoff 0
.mBar.file add command -label [::msgcat::mc "SaveAs..."] -underline 4 \
  -command [list text_saveas .win.mon .mBar.file]
.mBar.file add command -label [::msgcat::mc "Print..."] -underline 0 \
  -command [list printTextWidget .win.mon]
.mBar.file add separator
.mBar.file add command -label [::msgcat::mc "Exit"] -underline 1 \
  -command exit

# Create edit menu
.mBar add cascade -label [::msgcat::mc "Edit"] -underline 0 \
  -menu .mBar.edit
menu .mBar.edit -tearoff 0
.mBar.edit add command -label [::msgcat::mc "Copy"] -underline 0 \
  -state disabled -command [list tk_textCopy .win.mon]
.mBar.edit add separator
.mBar.edit add command -label [::msgcat::mc "Select all"] -underline 0 \
  -command [list text_selectAll .win.mon]

# Create help menu
.mBar add cascade -label [::msgcat::mc "Help"] -underline 0 \
  -menu .mBar.help
menu .mBar.help -tearoff 0
.mBar.help add command -label  [::msgcat::mc "Manual"] -underline 0 \
  -command [list openManual $Prj update]

# Create popup menu
menu .mPop -tearoff 0
.mPop add command -label [::msgcat::mc "Copy"] \
  -command [list tk_textCopy .win.mon] -state disabled
.mPop add separator
.mPop add command -label [::msgcat::mc "Select all"] \
  -command [list text_selectAll .win.mon]

# Set up list of copy menu entries
set CopyEntries {{.mBar.edit 0} {.mPop 0}}

# Create and pack frame window
ttk::frame .win
pack .win -expand yes -fill both

# Define message types
set MsgTypeList [list \
  [list I [::msgcat::mc "Invoked"]] \
  [list C [::msgcat::mc "Committed"]] \
  [list S [::msgcat::mc "Start"]] \
  [list F [::msgcat::mc "Finish"]] \
  [list W [::msgcat::mc "Warning"]] \
  [list E [::msgcat::mc "Error"]] \
  [list A [::msgcat::mc "Abort"]] \
  [list D [::msgcat::mc "?????"]] ]

# Define colouring tags
set TagList [list \
  [list BGgreen  "-background PaleGreen"] \
  [list BGgrey   "-background LightGrey"] \
  [list BGred    "-background #f63a20"] \
  [list BGyellow "-background Yellow"]]

# Tag selected message types
set TaggingList [list \
  [list F BGgrey] \
  [list C BGgreen] \
  [list E BGred] \
  [list A BGred] \
  [list W BGyellow]]

# Create monitor widget
set w [monitor_create .win.mon 83 24 $MsgTypeList $TagList \
  $TaggingList $CopyEntries .mPop]

# Get date
set Date [LocalNum [clock format [clock seconds] -format "%e %m %Y"]]
scan $Date "%s%s%s" day month year

# Format date
set Date [::msgcat::mc "(%1\$s/%2\$s/%3\$s)" $day $month $year]

# Build and show header
set fmt "Update of the %s Grade Model %s"
monitor_title .win.mon [::msgcat::mc $fmt $ProjectName $Date]

# Show usage if there are no arguments are given
if {$argc == 0} {
  set msg "No parameters specified (see usage note)"
  monitor_msg $w A [::msgcat::mc $msg]
  showUsage_update
  exit
}

# Read parameters for property estimation
if {! [getEstPars $PrjConfigDir EstPars ErrMsg]} {
  set msg "Configuration error (see message window)"
  monitor_msg $w A [::msgcat::mc $msg]
  set ErrTitle [::msgcat::mc "Configuration Error"]
  tk_messageBox -icon error -title $ErrTitle -message $ErrMsg
  exit
}

# Validate estimation parameters
if {! [validateEstPars $EstPars $RegionList $ModelAttrList ErrMsg]} {
  set msg "Incomplete or illegal estimation parameters"
  monitor_msg $w A [::msgcat::mc $msg]
  set ErrTitle [::msgcat::mc "Configuration Error"]
  tk_messageBox -icon error -title $ErrTitle -message $ErrMsg
  vwait BtnPress
}

# Initialise update parameter array
set Par(ProjectName) $PrjInfo(ProjectName)
set Par(DepositName) $PrjInfo(DepositName)
set Par(Resolution) $PrjInfo(Resolution)
set Par(GradeDir) $PrjInfo(GradeDir)
set Par(DataDir) $PrjInfo(DataDir)
set Par(GradeOctree) $PrjInfo(GradeOctree)
set Par(BackupOctree) "[file rootname $Par(GradeOctree)].bak"
set Par(ShadowOctree) "[file rootname $Par(GradeOctree)]temp.aob"
set Par(DataFileCols) $PrjInfo(DataFileCols)
set Par(Regs) $RegionList
set Par(Attrs) $ModelAttrList
set Par(EstPars) $EstPars

# Extract parameters from command-line argument vector
if [getUpdatePars argv argc Par $w] {
  
  # Perform model update
  if {[makeUpdate $Prj Par $w] == 1} {
    exit
  }
}
