########################################################################
#                                                                      #
# update               Tcl/Tk Procedures             File: update1.tcl #
#                                                                      #
# Tcl/Tk procedures to update selected mining regions in a linear      #
# octree encoded property model                                        #
#                                                                      #
# Date written  : 18/05/2000 by R. Prissang (FUB)                      #
# Last modified : 22/02/2015 (20:30) by R. Prissang (FUB)              #
#                                                                      #
########################################################################

package require msgcat
package require tdom

#-----------------------------------------------------------------------

proc showUsage_update {} {

#***********************************************************************
#
# Function
#
#     Creates a TextViewer window to show the usage description of
#     script update.tcl.
#
# Reqd. Procedures
#
#     text_append, TextViewer_create
#
# Date written  :  5/01/2011 by R. Prissang (FUB)
# Last modified : 11/06/2012 by R. Prissang (FUB)
#
#***********************************************************************

  # Return, if the text window is already open
  if [winfo exists .fv] {
    return
  }

  # Set title
  set Title "update (usage)"

  # Set tag text
  set TagList {{text "-justify left"}}

  # Create TextViewer window
  set TextWin [TextViewer_create .fv 80 40 $Title $TagList]

  # Show usage description
  text_append $TextWin "Usage:" text
  text_append $TextWin \
    "\n\nwish update.tcl \[partID ..\] \[Options ..\]" text
  text_append $TextWin \
    "\n\nFull or partial update of a grade octree" text

  # Show option description
  text_append $TextWin "\n\nOptions:\n" text
  set msg \
    "\n  -slices botIndex topIndex  range of slices in z direction\
     \n  -placard                   show placard as final message\n\
     \nOption -slices may be used to specify multiple ranges."
 
  # Append description of the options
  text_append $TextWin $msg text

  # Stay in the event loop and wait for the TextViewer being destroyed
  tkwait window .fv
}

#-----------------------------------------------------------------------

proc validateEstPars {EstParList RegionList AttrList ErrMsg} {

#***********************************************************************
#
# Function
#
#     Validates the selected regions and attributes.
#
# Arguments
#
#     RegionList : list of region parameters
#     AttrList   : attribute list
#     ErrMsg     : error message
#
# Return Value
#
#     1 : success
#     0 : failure
#
# Date written  : 10/03/2014 by R. Prissang (FUB)
# Last modified : 17/01/2015 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $ErrMsg msg

  # Initialise variable
  set msg ""

  # Check if the selected regions have been defined
  foreach Region $EstParList {
    if {[lsearch -index 0 $RegionList [lindex $Region 0]] < 0} {
      set fmt "Region \"%s\" had not been defined."
      set msg [::msgcat::mc $fmt [lindex $Region 0]]
      return 0
    }

    # Check if the selected attributes have been defined
    foreach Attr [lindex $Region 1] {
      if {[lsearch -index 0 $AttrList [lindex $Attr 0]] < 0} {
        set fmt "Attribute \"%s\" had not been defined."
        set msg [::msgcat::mc $fmt [lindex $Attr 0]]
        return 0
      }
    }
  }

  return 1
}

#-----------------------------------------------------------------------

proc getUpdatePars {argv argc ParArray w} {

#***********************************************************************
#
# Function
#
#     Inspects the argument vector to extract and check the control
#     parameters for a full or partial update of a grade octree.
#
#     The parameter set is expected to contain the following elements:
#
#     - numbers of parts of the model to be estimated
#     - options -slices bot top : indices of bottom and top slice
#               -placard        : show placard as a final message
#
# Arguments
#
#     argv     : argument vector
#     argc     : argument counter
#     ParArray : array of update parameters
#     w        : name of a monitor window
#
# Return Value
#
#     1 : success
#     0 : failure
#
# Reqd. Procedures
#
#     AsciiNum, getargs, LocalNum, monitor_msg
#
# Date written  :  2/01/2011 by R. Prissang (FUB)
# Last modified : 22/02/2015 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $argv Args
  upvar $argc nArgs
  upvar $ParArray Par

  # Check monitor widget
  if {! [winfo exists $w]} {
    error "Monitor widget $w does not exist"
  }

  # Initialise update parameter array
  set Par(Parts) {}
  set Par(Ranges) {}
  set Par(PlacardFlag) 0

  # Initialise argument description list
  set ArgDescrList [list {slices 2} {placard 0}]

  # Compute maximum slice index
  set Slice_max [expr (1 << $Par(Resolution))-1]

  # Return in case of an empty argument list
  if {[llength $Args] == 0} {
    set msg "Neither a region nor a slice specified"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Extract switches from the argument list
  set ReturnCode [getargs Args nArgs $ArgDescrList SwitchList]

  # Report errors, if any
  switch $ReturnCode {
    1 { # Extract illegal switches
      set s {}
      for {set i [expr $nArgs-1]} {$i >= 0} {incr i -1} {
        # Get argument
        set Arg [lindex $Args $i]

        # Skip floating point or integer values
        if [string is double $Arg] {continue}

        # Find switch
        if {[string index $Arg 0] eq "-"} {
          # Store illegal switch
          lappend s $Arg

          # Remove switch from argument list
          set Args [lreplace $Args $i $i]
        }
      }

      # Report illegal switches
      foreach Arg [lreverse $s] {
        set msg [::msgcat::mc "Option %s ignored" $Arg]
        monitor_msg $w W $msg
      }
    }

    2 { # Report that the parameters for switch -slices are incomplete
      set msg "Incomplete parameters for option -slices"
      monitor_msg $w E [::msgcat::mc $msg]
      return 0
    }
  }

  # Get and check switches and their parameters
  foreach s $SwitchList {
    # Set flag for showing a large success message window
    if {[string equal -nocase [lindex $s 0] "placard"] == 1} {
      set Par(PlacardFlag) 1
      continue
    }

    # Get ranges to be updated
    if {[string equal -nocase [lindex $s 0] "slices"] == 1} {
 
      # Get indices of the bottom and of the top slice
      set iBot [AsciiNum [lindex $s 1 0]]
      set iTop [AsciiNum [lindex $s 1 1]]

      # Check bottom slice
      if {! [string is integer $iBot]} {
        set fmt "Bottom slice index (%s) is not an integer"
        monitor_msg $w E [::msgcat::mc $fmt [lindex  $s 1 0]]
        return 0
      }
      if {$iBot < 0} {
        set fmt "Bottom slice index (%s) must be greater equal zero"
        monitor_msg $w E [::msgcat::mc $fmt [LocalNum $iBot]]
        return 0
      }
      if {$iBot > $Slice_max} {
        set fmt "Bottom slice index (%s) exceeding %s"
        monitor_msg $w E \
          [::msgcat::mc $fmt [LocalNum $iBot] [LocalNum $Slice_max]]
        return 0
      }

      # Check top slice
      if {! [string is integer $iTop]} {
        set fmt "Top slice index (%s) is not an integer"
        monitor_msg $w E [::msgcat::mc $fmt [lindex $s 1 1]]
        return 0
      }
      if {$iTop < 0} {
        set fmt "Top slice index (%s) must be greater equal zero"
        monitor_msg $w E [::msgcat::mc $fmt [LocalNum $iTop]]
        return 0
      }
      if {$iTop > $Slice_max} {
        set fmt "Top slice index (%s) exceeding %s"
        monitor_msg $w E \
          [::msgcat::mc $fmt [LocalNum $iTop] [LocalNum $Slice_max]]
        return 0
      }

      # Check range of slices
      if {$iBot > $iTop} {
        set fmt "Bottom slice (%s) above top slice (%s)"
        set msg [::msgcat::mc $fmt [LocalNum $iBot] [LocalNum $iTop]]
        monitor_msg $w E $msg
        return 0
      }

      # Remove sign, if necessary
      regsub {^\+} $iBot {} iBot
      regsub {^\+} $iTop {} iTop

      # Adjust index of slice zero
      regsub {^-?0+$} $iBot {0} iBot
      regsub {^-?0+$} $iTop {0} iTop

      # Append range
      set r [list $iBot $iTop]
      if {[lsearch -exact $Par(Ranges) $r] < 0} {
        lappend Par(Ranges) $r
      }
    }
  }

  # Sort list of ranges
  set Par(Ranges) [lsort -index 0 -integer $Par(Ranges)]

  # Get parts of the deposit to be updated
  foreach ID $Args {
    # Append ID of the part of the deposit, if not already stored
    if {[lsearch -exact $Par(Parts) $ID] < 0} {
      lappend Par(Parts) $ID
    }
  }

  # Sort list of parts of the deposit earmarked to be updated
  set Par(Parts) [lsort $Par(Parts)]

  # Return if neither parts of the deposit nor ranges are specified
  if {[llength $Par(Parts)] == 0 && [llength $Par(Ranges)] == 0} {
    set msg "Neither a region nor a slice specified"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  return 1
}

#-----------------------------------------------------------------------

proc makeUpdate {Prj ParArray w} {

#***********************************************************************
#
# Function
#
#     Performs an update of the grade model. The argument list may
#     contain any number of region numbers. If one of the arguments
#     contains the switch "-slice" or "-slices", a single slice
#     or the region between two slices will be updated. The 
#     appropriate parameter files are selected automatically.
#
# Arguments
#
#     Prj      : project code
#     ParArray : array of update parameters
#     w        : name of a monitor window
#
# Return Value
#
#      1 : success
#      0 : program error
#     -1 : abort
#
# Reqd. Procedures
#
#     showFinalMsg, updateRegions, updateSlices
#
# Date written  : 18/05/2000 by R. Prissang (FUB)
# Last modified :  7/03/2014 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $ParArray Par

  # Initialise variables
  set ProjectName [::msgcat::mc $Par(ProjectName)]
  set ImageFile "StBarbara.gif"

  # Return if no update parameters are given
  if {[llength $Par(Parts)] == 0 && [llength $Par(Ranges)] == 0} {
    return 0
  }

  # Check if parts of the deposit have been selected for updating
  if {[llength $Par(Parts)] > 0} {
    # Update selected parts of the deposit
    if {[updateRegions Par $w] <= 0} { 
      return -1
    }
  }

  # Check if slice ranges have been selected for updating
  if {[llength $Par(Ranges)] > 0} {
    # Update selected slices
    if {[updateSlices Par $w] <= 0} {
      return -1
    }
  }

  # Set up text strings for the region update completion window 
  set Title [::msgcat::mc "Completion Message"]
  set IconName [::msgcat::mc "Update OK"]

  if $Par(PlacardFlag) {
    set fmt "Update of the %s property model\nsuccessfully completed."
    set msg [::msgcat::mc $fmt $ProjectName]

    # Iconify monitor window
    wm iconify .

    # Display placard after successful update of the property model
    showFinalMsg .p $Prj $ImageFile $Title $IconName $msg
    
  } else {
    set msg "Update of the selected slices successfully\ncompleted."

    # Display message after successful update of the property model
    showFinalMsg .p $Prj -- $Title $IconName [::msgcat::mc $msg]
  }

  tkwait window .p
  return 1
}
