########################################################################
#                                                                      #
# update               Tcl/Tk Procedures             File: update2.tcl #
#                                                                      #
# Tcl/Tk procedures to update selected mining regions in a linear      #
# octree encoded property model                                        #
#                                                                      #
# Date written  :  3/05/2000 by R. Prissang (FUB)                      #
# Last modified : 24/04/2021 (22:00) by R. Prissang (FUB)              #
#                                                                      #
########################################################################

set ReturnCode ""

#-----------------------------------------------------------------------

proc updateRegions {ParArray w} {

#***********************************************************************
#
# Function
#
#     Updates selected regions in an octree encoded property model.
#
# Arguments
#
#     ParArray : array of update parameters
#     w        : name of a monitor window
#
# Return Value
#
#      1 : success
#      0 : failure
#
# Reqd. Procedures
#
#     buildEstSubModel, buildShadowGradeOctree, checkInputData, cleanup,
#     DirCheck, estimate, extractSubModel, fetchinp, getRegionName,
#     monitor_msg, TmpFileName, validateRegionIDs 
#
# Date written  :  6/05/2000 by R. Prissang (FUB)
# Last modified : 24/04/2021 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $ParArray Par

  # Check parameter array
  if {! [array exists Par]} {
    error "Variable Par is no array or it does not exist"
  }

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Initialise variables
  set GradeDir $Par(GradeDir)
  set GradeOctree "$GradeDir/$Par(GradeOctree)"
  set BackupOctree "$GradeDir/$Par(BackupOctree)"
  set ShadowOctree "$GradeDir/$Par(ShadowOctree)"

  # Validate IDs of the regions earmarked for updating
  if {! [validateRegionIDs Par $w]} {
    return 0
  }

  # Check, if the property model exists
  if {! [file exists $GradeOctree]} {
    set msg "Grade octree not found"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Check, if the grade octree may be modified
  if {! [file writable $GradeOctree]} {
    set msg "Grade octree is write protected"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Check, if the grade directory is writable
  if {! [file writable $GradeDir]} {
    set fmt "Directory %s is write protected"
    monitor_msg $w A [::msgcat::mc $fmt $GradeDir]
    return 0
  }

  # Make directory tmp, if not yet existing
  if {[DirCheck tmp 1] != 1} {
    set fmt "Directory %s can neither be created nor accessed"
    monitor_msg $w A [::msgcat::mc $fmt "tmp"]
    return 0
  }

  # Check input data
  if {[checkInputData Par $w] != 1} {
    return 0
  }

  # Update selected regions
  foreach i $Par(Parts) {
    # Get region name
    set RegionName [getRegionName $i $Par(Regs)]

    # Get region parameters
    set RegionPars [lsearch -index 0 -inline $Par(Regs) $i]

    # Display start message 
    monitor_msg $w I [::msgcat::mc "Update %s" $RegionName]

    # Create submodel
    set Par(SubMod) [TmpFileName submod${i}- aob tmp]

    # Check submodel
    if {$Par(SubMod) eq ""} {
      set msg "Submodel cannot be created"
      monitor_msg $w A [::msgcat::mc $msg]
      return 0
    }

    # Extract region
    if {[extractSubModel region $RegionPars $RegionName Par $w] <= 0} {
      return 0
    }

    # Retrieve data to estimate the properties in the mining region
    if {[fetchinp $i Par $w] != 1} {
      return 0
    }

    # Estimate properties in the current region
    foreach Prop $Par(Attrs) {
      # Estimate property
      if {[estimate [lindex $Prop 0] $i $RegionName Par Est $w] != 1} {
        return 0
      }
    }

    # Merge estimates of the attributes into a single octree
    if {[buildEstSubModel $i Est SubMod_est Par $w] != 1} {
      return 0
    }

    # Build shadow grade octree
    if {[buildShadowGradeOctree $SubMod_est Par $w] != 1} {
      return 0
    }

    # Backup existing grade octree
    if [catch {file copy -force $GradeOctree $BackupOctree}] {
      set msg "Backup of grade octree cannot be created"
      monitor_msg $w A [::msgcat::mc $msg]
      return 0
    }

    # Overwrite grade octree with its newer version
    if [catch {file rename -force $ShadowOctree $GradeOctree}] {
      set msg "Grade octree cannot be overwritten"
      monitor_msg $w A [::msgcat::mc $msg]
      return 0
    }

    # Remove backup of the grade octree
    if [catch {file delete $BackupOctree}] {
      set msg "Backup of grade octree cannot be deleted"
      monitor_msg $w A [::msgcat::mc $msg]
      return 0
    }

    # Display completion message in the monitor
    monitor_msg $w C [::msgcat::mc "Update %s" $RegionName]

    # Delete temporary files
    cleanup [pid]
  }

  return 1
}

#-----------------------------------------------------------------------

proc validateRegionIDs {ParArray w} {

#***********************************************************************
#
# Function
#
#     Validates the IDs of the regions to be estimated and sorts the
#     list ascendingly. Duplicate entries are removed.
#
# Arguments
#
#     ParArray : array of update parameters
#     w        : name of a monitor window
#
# Return Value
#
#     1 : success
#     0 : failure
#
# Reqd. Procedure
#
#     monitor_msg
#
# Date written  :  6/05/2000 by R. Prissang (FUB)
# Last modified : 17/01/2015 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $ParArray Par

  # Check parameter array
  if {! [array exists Par]} {
    error "Variable Par is no array or it does not exist"
  }

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Initialise variable
  set err 0

  # Count regions to be estimated
  set nRegions_est [llength $Par(Parts)]

  # Check number of regions to be estimated
  if {$nRegions_est == 0} {
    monitor_msg $w A [::msgcat::mc "No region specified"]
    return 0
  }

  # Check if the selected regions have been defined
  foreach Region $Par(Parts) {
    # Get region ID
    set RegionID [lindex $Region 0]

    # Search region list for current region ID
    if {[lsearch -index 0 $Par(Regs) $RegionID] < 0} {
      monitor_msg $w A [::msgcat::mc "Region %s unknown" $RegionID]
      set err 1
    }
  }

  # Return in case of an error
  if {$err == 1} {return 0}

  # Sort region IDs and remove duplicates
  if {$nRegions_est > 1} {
    # Sort region IDs ascendingly
    set Par(Parts) [lsort -dictionary $Par(Parts)]
 
    # Remove duplicate entries
    for {set i $nRegions_est} {$i > 0} {incr i -1} {
      if {[lindex $Par(Parts) $i] == [lindex $Par(Parts) [expr $i-1]]} {
        set Par(Parts) [lreplace $Par(Parts) $i $i]
      }
    }
  }

  return 1
}
