########################################################################
#                                                                      #
# update               Tcl/Tk Procedures             File: update3.tcl #
#                                                                      #
# Tcl/Tk procedures to update a given range of z-slices in an octree   #
# encoded property model                                               #
#                                                                      #
# Date written  :  3/05/2000 by R. Prissang (FUB)                      #
# Last modified : 24/04/2021 (22:00) by R. Prissang (FUB)              #
#                                                                      #
########################################################################

#-----------------------------------------------------------------------

proc updateSlices {ParArray w} {

#***********************************************************************
#
# Function
#
#     Updates a single slice or a range delineated by a bottom and
#     a top slice in an octree encoded property model.
#
# Arguments
#
#     ParArray : array of update parameters
#     w        : name of a monitor window
#
# Return Value
#
#      1 : success
#      0 : failure
#
# Reqd. Procedures
#
#     buildEstSubModel, buildShadowGradeOctree, checkInputData, cleanup,
#     DirCheck, estimate, extractSubModel, fetchinp, getRegionName,
#     monitor_msg, SliceList_create, TmpFileName
#
#  Date written  :  6/05/2000 by R. Prissang (FUB)
#  Last modified : 24/04/2021 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $ParArray Par

  # Check parameter array
  if {! [array exists Par]} {
    error "Variable Par is no array or it does not exist"
  }

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Initialise variables
  set GradeDir $Par(GradeDir)
  set GradeOctree "$GradeDir/$Par(GradeOctree)"
  set BackupOctree "$GradeDir/$Par(BackupOctree)"
  set ShadowOctree "$GradeDir/$Par(ShadowOctree)"
  set Slices $Par(Ranges)

  # Create slice list
  if {[SliceList_create $Slices $Par(Regs) SliceList $w] != 1} {
    return 0
  }

  # Check, if the property model exists
  if {! [file exists $GradeOctree]} {
    set msg "Grade octree not found"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Check, if the grade octree may be modified
  if {! [file writable $GradeOctree]} {
    set msg "Grade octree is write protected"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Check, if the grade directory is writable
  if {! [file writable $GradeDir]} {
    set fmt "Directory %s is write protected"
    monitor_msg $w A [::msgcat::mc $fmt $GradeDir]
    return 0
  }

  # Make directory tmp, if not yet existing
  if {[DirCheck tmp 1] != 1} {
    set fmt "Directory %s can neither be created nor accessed"
    monitor_msg $w A [::msgcat::mc $fmt "tmp"]
    return 0
  }

  # Check input data
  if {[checkInputData Par $w] != 1} {
    return 0
  }

  # Update slice ranges
  foreach i $SliceList {
    # Get ID of the region in which the slices are located
    set RegionID [lindex $i 2]

    # Get the name of this region
    set RegionName [getRegionName $RegionID $Par(Regs)]

    # Build slice name
    if {[lindex $i 0] != [lindex $i 1]} {
      set fmt "%s, slices %s-%s"
      set SliceName \
        [::msgcat::mc $fmt $RegionName [lindex $i 0] [lindex $i 1]]
    } else {
      set fmt "%s, slice %s"
      set SliceName \
        [::msgcat::mc $fmt $RegionName [lindex $i 0]]
    }

    # Display start message 
    monitor_msg $w I [::msgcat::mc "Update %s" $SliceName]

    # Create submodel
    set Par(SubMod) [TmpFileName submod${RegionID}- aob tmp]

    # Check submodel
    if {$Par(SubMod) eq ""} {
      set msg "Submodel cannot be created"
      monitor_msg $w A [::msgcat::mc $msg]
      return 0
    }

    # Extract slices 
    set ReturnCode [extractSubModel slices $i $SliceName Par $w]
    if {$ReturnCode <= 0} {
      return 0
    }

    # Retrieve data to estimate the properties in the given region
    set ReturnCode [fetchinp $RegionID Par $w]
    if {$ReturnCode != 1} {
      return 0
    }

    # Estimate properties
    foreach j $Par(Attrs) {
      # Get property ID
      set PropID [lindex $j 0]

      # Estimate property
      if {[estimate $PropID $RegionID $SliceName Par Est $w] != 1} {
        return 0
      }
    }

    # Merge estimates of all attributes into a single octree
    if {[buildEstSubModel $RegionID Est SubMod_est Par $w] != 1} {
      return 0
    }

    # Build shadow grade octree
    if {[buildShadowGradeOctree $SubMod_est Par $w] != 1} {
      return 0
    }

    # Backup existing grade octree
    if [catch {file copy -force $GradeOctree $BackupOctree}] {
      set msg "Backup of grade octree cannot be created"
      monitor_msg $w A [::msgcat::mc $msg]
      return 0
    }

    # Overwrite grade octree with its newer version
    if [catch {file rename -force $ShadowOctree $GradeOctree}] {
      set msg "Grade octree cannot be overwritten"
      monitor_msg $w A [::msgcat::mc $msg]
      return 0
    }

    # Remove backup of the grade octree
    if [catch {file delete $BackupOctree}] {
      set msg "Backup of grade octree cannot be deleted"
      monitor_msg $w A [::msgcat::mc $msg]
      return 0
    }

    # Display completion message in the monitor
    monitor_msg $w C [::msgcat::mc "Update %s" $SliceName]

    # Delete temporary files
    cleanup [pid]
  }

  return 1
}

#-----------------------------------------------------------------------

proc SliceList_create {Ranges RegionInfo SliceList w} {

#***********************************************************************
#
# Function
#
#     Sets up a list of slices to be updated. The resulting list
#     comprises elements containing the following information:
#
#       - Region ID
#       - bottom slice index
#       - top slice index
#
# Arguments
#
#     Ranges     : list of ranges to be updated
#     RegionInfo : list of region parameters
#     SliceList  : list of slices to be updated
#     w          : name of a monitor window
#
# Return Value
#
#     1 : success
#     0 : failure
#
# Reqd. Procedures
#
#     LocalNum, monitor_msg
#
# Date written  :  6/05/2000 by R. Prissang (FUB)
# Last modified : 16/01/2015 by R. Prissang (FUB)
#
#***********************************************************************

  upvar SliceList List

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Initialise variables
  set List {}
  set nRegions [llength $RegionInfo]

  # Return, in case no slice was selected
  if {[llength $Ranges] == 0} {
    monitor_msg $w A [::msgcat::mc "No slice specified"]
    return 0
  }

  # Inspect specified regions
  foreach Slices $Ranges {

    # Use slice index as bottom and as top, if only one slice is given
    if {[llength $Slices] < 2} {
      lappend Slices [lindex $Slices 0]
    }

    # Get bottom and the top slice indices
    set iBots [lindex $Slices 0]
    set iTops [lindex $Slices 1]

    # Check range
    if {$iBots > $iTops} {
      set fmt "Bottom slice (%s) above top slice (%s)"
      monitor_msg $w E \
        [::msgcat::mc $fmt [LocalNum $iBots] [LocalNum $iTops]]
      return 0
    }

    # Initialise variable    
    set done 0

    # Build slice list
    while {! $done} {
      for {set i 0} {$i < $nRegions} {incr i} {
        # Set variables
        set Region [lindex $RegionInfo $i]
        set RegionID [lindex $Region 0]
        set iBotr [lindex $Region 2]
        set iTopr [lindex $Region 3]

        # Check, if the bottom slice is located in the current region
        if {$iBotr <= $iBots && $iBots <= $iTopr} {
          # Check, if the top slice is located in the current region
          if {$iTops <= $iTopr} {
            # Append slice bottom, slice top & region ID to slice list
            lappend List [list $iBots $iTops $RegionID]
            set done 1
          } else {
            # Append slice bottom, region top & region ID to slice list
            lappend List [list $iBots $iTopr $RegionID]
            set iBots [expr $iTopr+1]
          }
          break
        }
      }

      # Display error message if bottom slice not included in any region
      if {[llength $List] == 0} {
        set fmt "Bottom slice (%s) not included in any region"
        monitor_msg $w E \
          [::msgcat::mc $fmt [LocalNum [lindex $Slices 0]]]
        return 0
      }

      # Display error message if top slice not included in any region
      if {$i == $nRegions} {
        set fmt "Top slice (%s) not included in any region"
        monitor_msg $w E \
          [::msgcat::mc $fmt [LocalNum [lindex $Slices 1]]]
        return 0
      }
    }
  }

  return 1
}
