########################################################################
#                                                                      #
# update               Tcl/Tk Procedures             File: update4.tcl #
#                                                                      #
# Tcl/Tk procedure to extract the obels to be estimated from an        #
# octree encoded property model                                        #
#                                                                      #
# Date written  : 14/05/2000 by R. Prissang (FUB)                      #
# Last modified :  8/08/2022 (12:30) by R. Prissang (FUB)              #
#                                                                      #
########################################################################

#-----------------------------------------------------------------------

proc extractSubModel {Type Descr Name ParArray w} {

#***********************************************************************
#
# Function
#
#     Front end for extracting a submodel from an octree. The
#     procedure runs program edito as a backend process.
#
# Argument
#
#     Type     : submodel type ("region" or "slices")
#     Descr    : submodel description
#     Name     : submodel name
#     ParArray : array of update parameters
#     w        : name of a monitor window
#
# Global Variable
#
#     ReturnCode
#
# Return Value
#
#      1 : success
#      0 : program error
#     -1 : abort
#
# Reqd. Procedures
#
#     countObels, fhandler, monitor_msg, text_loadFile,
#     TextViewer_create, TmpFileName
#
# Date written  : 14/05/2000 by R. Prissang (FUB)
# Last modified :  8/08/2022 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $ParArray Par

  global ReturnCode

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }
  
  # Initialise variables
  set GradeOctree "$Par(GradeDir)/$Par(GradeOctree)"
  set SubMod $Par(SubMod)
  set ReturnCode ""

  # Display start message
  monitor_msg $w S [::msgcat::mc "Extraction (%s)" $Name]

  # Get indices of the bottom and top slices of the submodel
  switch $Type {
    region { 
      set iBot [lindex $Descr 2]
      set iTop [lindex $Descr 3]
    }
    slices {
      set iBot [lindex $Descr 0]
      set iTop [lindex $Descr 1]
    }
    default {
      error "$Type: illegal submodel descriptor"
    }
  }

  # Check existence of template subex.par
  if {! [file exists par/subex.par]} {
    set fmt "Parameter file %s not found"
    monitor_msg $w A [::msgcat::mc $fmt "par/subex.par"]
    return -1
  }

  # Create temporary parameter file
  set TmpParFile [TmpFileName subex par tmp]

  # Verify existence of the temporary parameter file
  if {$TmpParFile eq ""} {
    set msg "Temporary parameter file could not be created"
    monitor_msg $w A [::msgcat::mc $msg] 
    return -1
  }

  # Open template subex.par  
  if [catch {set fd [open par/subex.par r]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt "par/subex.par"]
    return -1
  }

  # Read content of template subex.par
  set text [chan read $fd]

  # Close template subex.par
  catch {chan close $fd}

  # Apply changes to template subex.par
  regsub -all {DEPOSIT} $text $Par(ProjectName) text
  regsub -all {NAME} $text [regsub -all {&} $Name {\\\&}] text
  regsub {INFILE} $text \"$GradeOctree\" text
  regsub {OUTFILE} $text \"$SubMod\" text
  regsub {BOT} $text $iBot text
  regsub {TOP} $text $iTop text

  # Open temporary parameter file
  if [catch {set fd [open $TmpParFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
    return -1
  }

  # Set encoding scheme to UTF-8
  fconfigure $fd -encoding utf-8

  # Write temporary parameter file
  chan puts $fd $text

  # Close temporary parameter file
  catch {chan close $fd}

  # Check size of the temporary parameter file
  if {[file exists $TmpParFile] && [file size $TmpParFile] == 0} {
    set fmt "Parameter file %s incomplete"
    monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
    return -1
  }

  # Create log file
  set LogFile [TmpFileName log txt tmp]
  if {$LogFile eq ""} {
    set msg "Log file cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return -1
  }

  # Open log file
  if [catch {set fd_log [open $LogFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt $LogFile]
    return -1
  }

  # Extract a submodel
  set fd_backend [open "| edito $TmpParFile |& cat" "r+"]
  chan configure $fd_backend -buffering line

  # Set handler for a channel event
  chan event $fd_backend readable "fhandler $fd_backend $fd_log"

  # Stay in an event loop until the extraction process finishes
  vwait ReturnCode
 
  # Check log file
  if {[file exists $LogFile] && [file size $LogFile] == 0} {
    file delete $LogFile
    set msg "Log file could not be written"
    monitor_msg $w A [::msgcat::mc $msg]
    return -1
  }

  # Print error message, if necessary
  if {$ReturnCode != 0} {
    # Show error message
    set fmt "Error during submodel extraction"
    monitor_msg $w A [::msgcat::mc $fmt $Name]
    update
 
    # Show log file
    set Title [::msgcat::mc "Submodel Extraction"]
    set TagList {{text "-justify left"}}
    set TextWin [TextViewer_create .fv 82 40 $Title $TagList]
    text_loadFile $TextWin $LogFile text
    tkwait window .fv
    catch {file delete $LogFile}
    exit
  }

  # Delete log file
  catch {file delete $LogFile}

  # Display warning in case of an empty octree
  if {[countObels $SubMod] == 0} {
    set msg "Selected region does not overlap deposit model"
    monitor_msg $w W [::msgcat::mc $msg]
    return -1
  }

  # Display submodel extraction concluding message
  monitor_msg $w F [::msgcat::mc "Extraction (%s)" $Name]

  return 1
}
