########################################################################
#                                                                      #
# update               Tcl/Tk Procedures             File: update5.tcl #
#                                                                      #
# Tcl/Tk procedures to update selected mining regions in a linear      #
# octree encoded property model                                        #
#                                                                      #
# Date written  :  3/05/2000 by R. Prissang (FUB)                      #
# Last modified : 21/02/2022 (09:00) by R. Prissang (FUB)              #
#                                                                      #
########################################################################

set ReturnCode ""

#-----------------------------------------------------------------------

proc estimate {PropID RegionID RegionName ParArray Est w} {

#***********************************************************************
#
# Function
#
#     Estimates a selected property of the blocks in a given submodel
#     by means of stepwise interpolation.
#
# Arguments
#
#     PropID     : property ID
#     RegionID   : region ID
#     RegionName : region name 
#     ParArray   : array of update parameters
#     Est        : estimated submodel
#     w          : name of a monitor window
#
# Global Variable
#
#     ReturnCode
#
# Return Value
#
#      1 : success
#      0 : error
#     -1 : abort
#
# Reqd. Procedures
#
#     countObels, createOctrees, estoFrontEnd, extractUnknown,
#     GaugeWin_create, makeEstFiles, mergeEstimates, monitor_msg,
#     setMissVals
#
# Literature
#
#     PRISSANG, R. (1994) : The application of linear octree encoded
#       property modelling to the Cayeli deposit.- In: PASAMEHMETOGLU,
#       A.G., ESKIKAYA, S. KARPUZ, C. & HIZAL, T. (1994): Mine Planning
#       and Equipment Selection'94.- 557-562, (Balkema),
#       Rotterdam-Brookefield
#
# Date written  :  6/05/2000 by R. Prissang (FUB)
# Last modified : 22/02/2015 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $ParArray Par
  upvar $Est Estimates 

  global ReturnCode, monitorInfo, MsgType

  # Check property ID
  if {[string trim $PropID] eq ""} {
    error "No property ID given"
  }

  # Check region ID
  if {[string trim $RegionID] eq ""} {
    error "No region ID given"
  }

  # Check parameter array
  if {! [array exists Par]} {
    error "Variable Par is no array or it does not exist"
  }

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Verify existence of a submodel containing the obels to be estimated
  if {! [file exists $Par(SubMod)]} {
    set msg "Submodel of the region to be estimated not found"
    monitor_msg $w A [::msgcat::mc $msg]
    return -1
  }

  # Count obels in the submodel
  set nObels [countObels $Par(SubMod)]

  # Return if the submodel is empty
  if {$nObels <= 0} {
    set msg "Selected region does not overlap deposit model"
    monitor_msg $w A [::msgcat::mc $msg]
    return -1
  }

  # Get property index
  set iProp [lsearch -index 0 $Par(Attrs) $PropID]

  # Check property index
  if {$iProp < 0} {
    set fmt "Property ID \"%s\" unknown"
    monitor_msg $w A [::msgcat::mc $fmt $PropID]
    return -1
  }

  # Get property name
  set PropName [lindex $Par(Attrs) $iProp 1]

  # Check region ID
  if {[lsearch -index 0 $Par(Regs) $RegionID] < 0} {
    set fmt "Region ID \"%s\" unknown"
    monitor_msg $w A [::msgcat::mc $fmt $RegionID]
    return -1
  }

  # Create octrees required for the estimation of the given property
  if {[createOctrees $PropID $RegionID Octree $w] == 0} {
    return -1
  }

  # Get path to the data file
  set DataFile $Par($RegionID,$PropID,dat)

  # Check existence of the data file
  if {! [file exists $DataFile]} {
    set fmt "Data file %s not found"
    monitor_msg $w A [::msgcat::mc $fmt $DataFile]
    return -1
  }

  # Set attribute to missing value in case of an empty data file
  if {[file size $DataFile] == 0} {
    # Show start message
    set msg [::msgcat::mc "Setting %s to missing value" $PropName]
    monitor_msg $w S $msg

    # Set attribute values to missing value
    if {[setMissVals $PropID $RegionID Octree Par $w] != 1} {
      return -1
    }

    # Store path to the estimates
    set Estimates($RegionID,$PropID) $Octree($RegionID,$PropID,est0)

    # Show completion message and return
    monitor_msg $w F $msg
    return 1
  }

  # Get list of estimation parameters for the given region
  set EstPars [lsearch -index 0 -inline $Par(EstPars) $RegionID]

  # Get list of estimation parameters for the given property
  set EstPars [lsearch -index 0 -inline [lindex $EstPars 1] $PropID]

  # Check list of estimation parameters
  if {[llength $EstPars] == 0} {
    set fmt "No estimation parameters for %s given"
    monitor_msg $w A [::msgcat::mc $fmt $PropName]
    return -1
  }

  # Check list of estimation parameters
  if {[llength $EstPars] == 0} {
    set fmt "No estimation parameters given (%s in region %s)"
    monitor_msg $w A [::msgcat::mc $fmt $PropName $RegionID]
    return -1
  }

  # Extract the list of estimation parameter files
  set ParSetList [lindex $EstPars 1]

  # Get number of steps
  set nSteps [llength $ParSetList]

  # Check list of estimation parameters
  if {$nSteps == 0} {
    set fmt "No estimation parameters given (%s in region %s)"
    monitor_msg $w A [::msgcat::mc $fmt $PropName $RegionID]
    return -1
  }

  # Build array of temporary parameter files and temporary octrees
  if {[makeEstFiles $PropID $RegionID Par ParFile Octree $w] != 1} {
    return -1
  }

  # Run first step of the property prediction process
  set i 0

  # Get parameter set
  set ParSet [lindex $ParSetList $i]

  # Get name of the estimation technique
  set EstTechn [::msgcat::mc [lindex $ParSet 0]]

  # Set up task description
  set fmt "%s (%s, %s)"
  set TaskName [::msgcat::mc $fmt $EstTechn $RegionName $PropName]

  # Set up workload description
  set fmt "%s blocks"
  set Workload [::msgcat::mc $fmt [LocalNum $nObels]]

  # Create and initialise progress gauge
  set Gauge [GaugeWin_create .gt 0 300 $TaskName $Workload \
    "PaleGreen" [::msgcat::mc "Update Progress"]]

  # Set index of the parameter file
  set Index "$RegionID,$PropID,est$i"

  # Estimate obels
  set ReturnCode [estoFrontEnd $w $ParFile($Index) $TaskName $Gauge]

  # Destroy gauge
  destroy .gt

  # Return in case of an error
  if {$ReturnCode != 1} {
    return $ReturnCode
  }

  # Run subsequent steps of the property prediction process
  for {set i 1} {$i < $nSteps} {incr i} {
    # Get parameter set
    set ParSet [lindex $ParSetList $i]

    # Set index of the parameter file
    set Index "$RegionID,$PropID,ex[expr $i > 1]"

    # Extract unknown obels from a previously estimated octree
    set ReturnCode [extractUnknown $ParFile($Index) $w]
    if {$ReturnCode != 1} {
      return $ReturnCode
    }

    # Count unknown obels
    set nObels [countObels $Octree($RegionID,$PropID,miss)]

    # Break, if all obels have been estimated
    if {$nObels == 0} {
      break
    }

    # Get name of the estimation technique
    set EstTechn [::msgcat::mc [lindex $ParSet 0]]

    # Set up task description
    set fmt "%s (%s, %s)"
    set TaskName [::msgcat::mc $fmt $EstTechn $RegionName $PropName]

    # Set up workload description
    set fmt "%s blocks"
    set Workload [::msgcat::mc $fmt [LocalNum $nObels]]

    # Create and initialise progress gauge
    set Gauge [GaugeWin_create .gt 0 300 $TaskName $Workload \
      "PaleGreen" [::msgcat::mc "Update Progress"]]

    # Set index of the parameter file
    set Index "$RegionID,$PropID,est$i"

    # Estimate obels
    set ReturnCode [estoFrontEnd $w $ParFile($Index) $TaskName $Gauge]

    # Destroy gauge
    destroy .gt

    # Return in case of an error
    if {$ReturnCode != 1} {
      return $ReturnCode
    }

    # Merge estimation results
    set ReturnCode [mergeEstimates $PropID $RegionID Par Octree $w]
    if {$ReturnCode != 1} {
      return $ReturnCode
    }
  }

  # Store path to the estimates
  set Estimates($RegionID,$PropID) $Octree($RegionID,$PropID,est0)

  # Delete temporary octrees
  catch {file delete $Octree($RegionID,$PropID,est1)}
  catch {file delete $Octree($RegionID,$PropID,miss)}

  return 1
}

#-----------------------------------------------------------------------

proc createOctrees {PropID RegionID OctArray w} {

#***********************************************************************
#
# Function
#
#     Sets up the temporary octrees for the estimation of a given
#     property. The names of the temporary octrees are stored in an
#     associative array.
#
# Arguments
#
#     PropID   : property ID
#     RegionID : region ID
#     OctArray : array of temporary octrees
#     w        : name of a monitor window
#
# Global Variable
#
#     ReturnCode
#
# Return Value
#
#     0 : success
#     1 : failure
#
# Reqd. Procedures
#
#     monitor_msg, TmpFileName
#
# Date written  : 25/06/2012 by R. Prissang (FUB)
# Last modified : 22/01/2015 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $OctArray Octree

  # Check property ID
  if {[string trim $PropID] eq ""} {
    error "No property ID given"
  }

  # Check region ID
  if {[string trim $RegionID] eq ""} {
    error "No region ID given"
  }

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Create octree to store the results of the first estimation run
  set TmpOctree [TmpFileName "$PropID${RegionID}est0-" aob tmp]

  # Check octree to store the results of the first estimation run
  if {$TmpOctree eq ""} {
    set msg "Primary estimation octree cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Store octree path
  set Octree($RegionID,$PropID,est0) $TmpOctree

  # Create octree to store the results of a subsequent estimation run
  set TmpOctree [TmpFileName "$PropID${RegionID}est1-" aob tmp]

  # Check octree to store the results of a subsequent estimation run
  if {$TmpOctree eq ""} {
    set msg "Secondary estimation octree cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Store octree path
  set Octree($RegionID,$PropID,est1) $TmpOctree

  # Create octree to store the unknown obels
  set TmpOctree [TmpFileName $PropID${RegionID}miss aob tmp]

  # Check octree to store the unknown obels
  if {$TmpOctree eq ""} {
    set msg "Octree of unknown blocks cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Store octree path
  set Octree($RegionID,$PropID,miss) $TmpOctree

  return 1
}

#-----------------------------------------------------------------------

proc setMissVals {PropID RegionID OctArray ParArray w} {

#***********************************************************************
#
# Function
#
#     Sets the value of a given property in all obels of a submodel to
#     missing value. The procedure runs program edito as a backend
#     process. 
#
# Arguments
#
#     PropID   : property ID
#     RegionID : region ID
#     OctArray : array of temporary octrees
#     ParArray : array of update parameters
#     w        : name of a monitor window
#
#
# Global Variable
#
#     ReturnCode
#
# Return Value
#
#     1 : success
#     0 : failure
#
# Reqd. Procedures
#
#     fhandler, getRegionName, monitor_msg, text_loadFile,
#     TextViewer_create, TmpFileName
#
# Date written  :  5/06/2012 by R. Prissang (FUB)
# Last modified :  8/04/2020 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $OctArray Octree
  upvar $ParArray Par

  global ReturnCode

  # Check property ID
  if {[string trim $PropID] eq ""} {
    error "No property ID given"
  }

  # Check region ID
  if {[string trim $RegionID] eq ""} {
    error "No region ID given"
  }

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Get property index
  set iProp [lsearch -index 0 $Par(Attrs) $PropID]

  # Check property index
  if {$iProp < 0} {
    set fmt "Property ID \"%s\" unknown"
    monitor_msg $w A [::msgcat::mc $fmt $PropID]
    return 0
  }

  # Get property name
  set PropName [lindex $Par(Attrs) $iProp 1]

  # Check region ID
  if {[lsearch -index 0 $Par(Regs) $RegionID] < 0} {
    set fmt "Region ID \"%s\" unknown"
    monitor_msg $w A [::msgcat::mc $fmt $RegionID]
    return 0
  }

  # Get region name
  set RegionName [getRegionName $RegionID $Par(Regs)]

  # Check existence of template setmiss.par
  if {! [file exists par/setmiss.par]} {
    set fmt "Parameter file %s not found"
    monitor_msg $w A [::msgcat::mc $fmt "setmiss.par"]
    return 0
  }

  # Create parameter file for adding a missing value
  set TmpParFile [TmpFileName "$PropID${RegionID}miss" par tmp]

  # Verify existence of the temporary parameter file
  if {$TmpParFile eq ""} {
    set msg "Temporary parameter file could not be created"
    monitor_msg $w A [::msgcat::mc $msg] 
    return 0
  }

  # Open template setmiss.par  
  if [catch {set fd [open par/setmiss.par r]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt "par/setmiss.par"]
    return 0
  }

  # Read content of template setmiss.par
  set text [chan read $fd]

  # Close template setmiss.par
  catch {chan close $fd}

  # Apply changes to template setmiss.par
  regsub -all {DEPOSIT} $text $Par(ProjectName) text
  regsub -all {REGION} $text $RegionName text
  regsub -all {ATTRIBUTE} $text $PropName text
  regsub {INFILE} $text \"$Par(SubMod)\" text
  regsub {OUTFILE} $text \"$Octree($RegionID,$PropID,est0)\" text

  # Open temporary parameter file
  if [catch {set fd [open $TmpParFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
    return 0
  }

  # Write temporary parameter file
  chan puts $fd $text

  # Close temporary parameter file
  catch {chan close $fd}

  # Check size of the temporary parameter file
  if {[file exists $TmpParFile] && [file size $TmpParFile] == 0} {
    set fmt "Parameter file %s incomplete"
    monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
    return 0
  }

  # Create log file
  set LogFile [TmpFileName log txt tmp]

  # Check log file
  if {$LogFile eq ""} {
    set msg "Log file cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Open log file
  if [catch {set fd_log [open $LogFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt $LogFile]
    return 0
  }

  # Set attribute values to missing value
  set fd_backend [open "| edito $TmpParFile |& cat" "r+"]
  chan configure $fd_backend -buffering line

  # Set handler for a channel event
  chan event $fd_backend readable "fhandler $fd_backend $fd_log"

  # Stay in an event loop until the attribute values are set
  vwait ReturnCode
 
  # Check log file
  if {[file exists $LogFile] && [file size $LogFile] == 0} {
    file delete $LogFile
    set msg "Log file could not be written"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Print error message, if necessary
  if {$ReturnCode != 0} {
    # Show error message
    set msg "Error during missing value assignment"
    monitor_msg $w A [::msgcat::mc $msg]
    update
 
    # Show log file
    set Title [::msgcat::mc "Missing Value Assignment"]
    set TagList {{text "-justify left"}}
    set TextWin [TextViewer_create .fv 82 40 $Title $TagList]
    text_loadFile $TextWin $LogFile text
    tkwait window .fv
    catch {file delete $LogFile}
    exit
  }

  # Delete log file
  catch {file delete $LogFile}

  return 1
}

#-----------------------------------------------------------------------
 
proc makeEstFiles {PropID RegionID ParArray ParFileArray OctArray w} {

#***********************************************************************
#
# Function
#
#     Sets up the parameter files for the estimation of a given
#     property. The names of the temporary parameter files are stored
#     in an associative array.
#
# Arguments
#
#     PropID       : property ID
#     RegionID     : region ID
#     ParArray     : array of update parameters
#     ParFileArray : array of temporary parameter files
#     OctArray     : array of temporary octrees
#     w            : name of a monitor window
#
# Return Value
#
#     1 : success
#     0 : failure
#
# Reqd. Routines
#
#     monitor_msg, TmpFileName
#
# Date written  : 13/05/2000 by R. Prissang (FUB)
# Last modified : 21/02/2022 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $ParArray Par
  upvar $ParFileArray ParFile
  upvar $OctArray Octree

  # Check property ID
  if {[string trim $PropID] eq ""} {
    error "No property ID given"
  }

  # Check region ID
  if {[string trim $RegionID] eq ""} {
    error "No region ID given"
  }

  # Check parameter array
  if {! [array exists Par]} {
    error "Variable Par is no array or it does not exist"
  }

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Initialise variables
  set nCols [llength $Par(DataFileCols)]

  # Get property index
  set iProp [lsearch -index 0 $Par(Attrs) $PropID]

  # Check property index
  if {$iProp < 0} {
    set fmt "Property ID \"%s\" unknown"
    monitor_msg $w A [::msgcat::mc $fmt $PropID]
    return 0
  }

  # Get property name
  set PropName [lindex $Par(Attrs) $iProp 1]

  # Get list of estimation parameters for the given region
  set EstPars [lsearch -index 0 -inline $Par(EstPars) $RegionID]

  # Get list of estimation parameters for the given property
  set EstPars [lsearch -index 0 -inline [lindex $EstPars 1] $PropID]

  # Check list of estimation parameter
  if {[llength $EstPars] == 0} {
    set fmt "No estimation parameters given (%s in region %s)"
    monitor_msg $w A [::msgcat::mc $fmt $PropName $RegionID]
    return 0
  }

  # Extract list of estimation parameter files
  set ParSetList [lindex $EstPars 1]

  # Get number of steps
  set nSteps [llength $ParSetList]

  # Check list of estimation parameters
  if {$nSteps == 0} {
    set fmt "No estimation parameters given (%s in region %s)"
    monitor_msg $w A [::msgcat::mc $fmt $PropName $RegionID]
    return 0
  }

  # Get position of the z coordinate in the project data file
  set iz [lsearch -index 0 $Par(DataFileCols) "z"]

  # Build format to read the coordinates
  for {set i 0} {$i <= $iz} {incr i} {
    if {[lindex $Par(DataFileCols) $i 0] in {x y z}} {
      append fmtc "%lf"
    } else {
      append fmtc "%*s"
    }
  }

  # Build format to read the attribute values
  for {} {$i < $nCols} {incr i} {
    if {[lindex $Par(DataFileCols) $i 0] eq $PropID} {
      append fmta "%f"
    } else {
      append fmta "%*s"
    }
  }

  # Build parameter files for each step in the prediction process
  for {set i 0} {$i < $nSteps} {incr i} {
    # Get parameter set
    set ParSet [lindex $ParSetList $i]

    # Get parameter file
    set ParFileName [lindex $ParSet 1]

    # Build path to the parameter file
    set Path "par/$ParFileName"

    # Check existence of the template with estimation control parameters
    if {! [file exists $Path]} {
      set fmt "Parameter file %s not found"
      monitor_msg $w A [::msgcat::mc $fmt $Path]
      return 0
    }

    # Create temporary parameter file
    set TmpParFile [TmpFileName "${PropID}${RegionID}est$i-" par tmp]

    # Verify existence of the temporary parameter file
    if {$TmpParFile eq ""} {
      set msg "Temporary parameter file could not be created"
      monitor_msg $w A [::msgcat::mc $msg] 
      return 0
    }

    # Set submodel to be estimated
    if {$i == 0} {
      set SubMod $Par(SubMod)
    } else {
      set SubMod $Octree($RegionID,$PropID,miss)
    }

    # Open template with estimation control parameters
    if [catch {set fd [open $Path r]}] {
      set fmt "Can't open file %s"
      monitor_msg $w A [::msgcat::mc $fmt $Path]
      return 0
    }

    # Read estimation control parameters
    set text [chan read $fd]

    # Close template with estimation control parameters
    catch {chan close $fd}
#BASTEL
    # Change estimation control parameters
    regsub {INFILE} $text \"$SubMod\" text
    regsub {DATAFILE} $text \"$Par($RegionID,$PropID,dat)\" text
    regsub {OUTFILE} $text \
      \"$Octree($RegionID,$PropID,est[expr $i > 0])\" text
    regsub {FMTCOOR} $text $fmtc text
    regsub {FMTATTR} $text $fmta text

    # Open temporary parameter file
    if [catch {set fd [open $TmpParFile w]}] {
      set fmt "Can't open file %s"
      monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
      return 0
    }

    # Write temporary parameter file
    chan puts $fd $text

    # Close temporary parameter file
    catch {chan close $fd}

    # Check size of the temporary parameter file
    if {[file exists $TmpParFile] && [file size $TmpParFile] == 0} {
      set fmt "Parameter file %s incomplete"
      monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
      return 0
    }

    # Store name of the temporary parameter file
    set ParFile($RegionID,$PropID,est$i) $TmpParFile
  }

  # Check number of steps in the estimation process
  if {$nSteps > 1} {
    # Check existence of template missex.par
    if {! [file exists par/missex.par]} {
      set fmt "Parameter file %s not found"
      monitor_msg $w A [::msgcat::mc $fmt missex.par]
      return 0
    }

    # Build temporary parameter files to extract unknown blocks
    for {set i 0} {$i <= 1} {incr i} {
      # Create temporary parameter file
      set TmpParFile [TmpFileName $PropID${RegionID}mex$i- par tmp]

      # Verify existence of the temporary parameter file
      if {$TmpParFile eq ""} {
        set msg "Temporary parameter file could not be created"
        monitor_msg $w A [::msgcat::mc $msg] 
        return 0
      }

      # Open template missex.par
      if [catch {set fd [open par/missex.par r]}] {
        set fmt "Can't open file %s"
        monitor_msg $w A [::msgcat::mc $fmt "par/missex.par"]
        return 0
      }

      # Read template missex.par
      set text [chan read $fd]

      # Close template missex.par
      catch {chan close $fd}

      # Apply changes to template missex.par
      regsub -all {DEPOSIT} $text $Par(ProjectName) text
      regsub {INFILE} $text \
        \"$Octree($RegionID,$PropID,est[expr $i > 0])\" text
      regsub {OUTFILE} $text \"$Octree($RegionID,$PropID,miss)\" text

      # Open temporary parameter file
      if [catch {set fd [open $TmpParFile w]}] {
        set fmt "Can't open file %s"
        monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
        return 0
      }

      # Write temporary parameter file
      chan puts $fd $text

      # Close temporary parameter file
      catch {chan close $fd}

      # Check size of the temporary parameter file
      if {[file exists $TmpParFile] && [file size $TmpParFile] == 0} {
        set fmt "Parameter file %s incomplete"
        monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
        return 0
      }

      # Store name of the temporary parameter file
      set ParFile($RegionID,$PropID,ex$i) $TmpParFile
    }
  }

  return 1
}

#-----------------------------------------------------------------------

proc mergeEstimates {PropID RegionID ParArray OctArray w} {

#***********************************************************************
#
# Function
#
#     Front end for merging the predictions for a given attribute
#     resulting from two subsequent estimation runs.
#
#     The procedure runs program calculo as a backend process.
#
# Arguments
#
#     PropID   : property ID
#     RegionID : region ID
#     ParArray : array of update parameters
#     OctArray : array of octrees
#     w        : name of a monitor window
#
# Global Variable
#
#     ReturnCode
#
# Return Value
#
#      1 : success
#      0 : failure
#
# Reqd. Procedures
#
#     fhandler, monitor_msg, text_loadFile, TextViewer_create,
#     TmpFileName
#
# Date written  : 22/06/2012 by R. Prissang (FUB)
# Last modified : 12/01/2015 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $OctArray Octree
  upvar $ParArray Par

  global ReturnCode

  # Check property ID
  if {[string trim $PropID] eq ""} {
    error "No property ID given"
  }

  # Check region ID
  if {[string trim $RegionID] eq ""} {
    error "No region ID given"
  }

  # Check parameter array
  if {! [array exists Par]} {
    error "Variable Par is no array or it does not exist"
  }

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Check monitor widget
  if {! [winfo exists $w]} {
    error "Monitor widget $w does not exist"
  }

  # Initialise variable
  set nAttrs 1

  # Get list of estimation parameters for the given region
  set EstParsReg [lsearch -index 0 -inline $Par(EstPars) $RegionID]

  # Get list of estimation parameters for the given property
  set EstPars [lsearch -index 0 -inline [lindex $EstParsReg 1] $PropID]

  # Get property name
  set PropName [lindex [lsearch -index 0 -inline $Par(Attrs) $PropID] 1]

  # Get names of primary and secondary estimation octrees
  set Oct0 $Octree($RegionID,$PropID,est0)
  set Oct1 $Octree($RegionID,$PropID,est1)

  # Check primary octree
  if {! [file exists $Oct0]} {
    set fmt "Octree %s cannot be found"
    monitor_msg $w A [::msgcat::mc $fmt $Oct0]
    return 0
  }

  # Check secondary octree
  if {! [file exists $Oct1]} {
    set fmt "Octree %s cannot be found"
    monitor_msg $w A [::msgcat::mc $fmt $Oct1]
    return 0
  }

  # Create octree to store the combined estimation results
  set Oct2 [TmpFileName "$PropID${RegionID}est2-" aob tmp]

  # Check octree to store the combined estimation results
  if {$Oct2 eq ""} {
    set msg "Temporary octree cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Check existence of template merge.par
  if {! [file exists "par/merge.par"]} {
    set fmt "Parameter file %s not found"
    monitor_msg $w A [::msgcat::mc $fmt "merge.par"]
    return 0
  }

  # Create temporary parameter file
  set TmpParFile [TmpFileName "$PropID${RegionID}merge" par tmp]

  # Verify existence of the temporary parameter file
  if {$TmpParFile eq ""} {
    set msg "Temporary parameter file could not be created"
    monitor_msg $w A [::msgcat::mc $msg] 
    return 0
  }

  # Open template merge.par
  if [catch {set fd [open par/merge.par r]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt "par/merge.par"]
    return 0
  }

  # Read template merge.par
  set text [chan read $fd]

  # Close template merge.par
  catch {chan close $fd}

  # Apply changes to the template for merging the estimates
  regsub -all {DEPOSIT} $text $Par(ProjectName) text
  regsub -all {PROPERTY} $text $PropName text
  regsub {FILE0} $text \"$Oct0\" text
  regsub {FILE1} $text \"$Oct1\" text
  regsub {OUTFILE} $text \"$Oct2\" text
  regsub {NATTRS} $text [format "%-6d" $nAttrs] text
#
#
#
#
# HIER KNNTEN WEITER ATTRIBUTE GESETZT WERDEN
# falls in der attributliste (DistCl krigingVar)
#
# Wird die Datei mehrmals bentigt?




  # Open temporary parameter file
  if [catch {set fd [open $TmpParFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
    return 0
  }

  # Write temporary parameter file
  chan puts $fd $text

  # Close temporary parameter file
  catch {chan close $fd}

  # Check size of the temporary parameter file
  if {[file exists $TmpParFile] && [file size $TmpParFile] == 0} {
    set fmt "Parameter file %s incomplete"
    monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
    return 0
  }

  # Create log file
  set LogFile [TmpFileName log txt tmp]

  # Check log file
  if {$LogFile eq ""} {
    set msg "Log file cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Open log file
  if [catch {set fd_log [open $LogFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt $LogFile]
    return 0
  }

  # Merge estimation results
  set fd_backend [open "| calculo $TmpParFile |& cat" "r+"]
  chan configure $fd_backend -buffering line

  # Set handler for a channel event
  chan event $fd_backend readable "fhandler $fd_backend $fd_log"

  # Stay in an event loop until the update process finishes
  vwait ReturnCode
 
  # Check log file
  if {[file exists $LogFile] && [file size $LogFile] == 0} {
    file delete $LogFile
    set msg "Log file could not be written"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Print error message, if necessary
  if {$ReturnCode != 0} {
    # Show error message
    set msg "Error during merging of the estimates"
    monitor_msg $w A [::msgcat::mc $msg]
    update
 
    # Show log file
    set Title [::msgcat::mc "Merging of Estimates"]
    set TagList {{text "-justify left"}}
    set TextWin [TextViewer_create .fv 82 40 $Title $TagList]
    text_loadFile $TextWin $LogFile text
    tkwait window .fv
    catch {file delete $LogFile}
    exit
  }

  # Rename output octree
  if [catch {file rename -force $Oct2 $Oct0}] {
    set fmt "Octree %s cannot be overwritten"
    monitor_msg $w A [::msgcat::mc $fmt $Oct0]
    return 0
  }

  # Delete log file
  catch {file delete $LogFile}

  return 1
}

#-----------------------------------------------------------------------

proc buildEstSubModel {RegionID EstOct SubModel ParArray w} {

#***********************************************************************
#
# Function
#
#     Front end to build the estimated submodel by merging the octrees
#     containing the estimates of the individual properties.
#
#     The procedure runs program calculo as a backend process.
#
# Arguments
#
#     RegioID  : region ID
#     EstOct   : array of estimated submodels
#     SubMod   : estimated submodel
#     ParArray : array of update parameters
#     w        : name of a monitor window
#
# Global Variable
#
#     ReturnCode
#
# Return Value
#
#      1 : success
#      0 : failure
#
# Reqd. Procedures
#
#     fhandler, monitor_msg, text_loadFile, TextViewer_create,
#     TmpFileName
#
# Date written  : 14/05/2000 by R. Prissang (FUB)
# Last modified :  9/04/2020 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $EstOct EstOctree
  upvar $SubModel SubMod
  upvar $ParArray Par

  global ReturnCode

  # Check region ID
  if {[string trim $RegionID] eq ""} {
    error "No region ID given"
  }

  # Check parameter array
  if {! [array exists Par]} {
    error "Variable Par is no array or it does not exist"
  }

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Initialise variables
  set nAttrs [llength $Par(Attrs)]
  set TaskName [::msgcat::mc "Set-up of estimated submodel"]

  # Display start message
  monitor_msg $w S $TaskName

  # Create temporary octree to store the estimated submodel
  set TmpOctree [TmpFileName "estmod${RegionID}-" aob tmp]

  # Check octree to store the estimated submodel
  if {$TmpOctree eq ""} {
    set msg "Estimated submodel cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Store octree path
  set SubMod $TmpOctree

  # Check existence of template mergeall.par
  if {! [file exists par/mergeall.par]} {
    set fmt "Parameter file %s not found"
    monitor_msg $w A [::msgcat::mc $fmt "mergeall.par"]
    return 0
  }

  # Create temporary parameter file
  set TmpParFile [TmpFileName "mergeall" par tmp]

  # Verify existence of the temporary parameter file
  if {$TmpParFile eq ""} {
    set msg "Temporary parameter file could not be created"
    monitor_msg $w A [::msgcat::mc $msg] 
    return 0
  }

  # Open template mergeall.par
  if [catch {set fd [open par/mergeall.par r]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt "par/mergeall.par"]
    return 0
  }

  # Read template mergeall.par
  set text [chan read $fd]

  # Close template mergeall.par
  catch {chan close $fd}

  # Apply changes to template mergeall.par
  regsub -all {DEPOSIT} $text $Par(ProjectName) text
  regsub {NFILES} $text [format "%-6s" $nAttrs] text

  # Build input file section for template mergeall.par
  set i 0
  foreach a $Par(Attrs) {
    # Append input file
    if {$i > 0} {append InFiles "\n"}
    append InFiles "# [lindex $a 1] estimates"
    append InFiles "\n\"$EstOctree($RegionID,[lindex $a 0])\""

    # Increment attribute counter
    incr i
  }

  # Insert input file section in template mergeall.par
  regsub {INFILES} $text $InFiles text

  # Set output file and number of attributes
  regsub {OUTFILE} $text \"$SubMod\" text
  regsub {NATTRS} $text [format "%-6s" $nAttrs] text

  # Set format strings for appending attributes descriptions
  set fmt0 "%-21s # Attribute name (max. 80 characters)"
  set fmt1 "%-21d # Attribute type (2=>FLT, 1=INT, 0=ENUM)"
  set fmt2 "%-21s # VSL Expression to be evaluated"

  # Add attribute descriptions to template mergeall.par
  set i 0
  foreach a $Par(Attrs) {
    # Get attribute ID, name, and name supplement
    lassign $a AttrID Name Supplement

    # Set attribute name
    if {[string trim $Supplement] ne ""} {
      set AttrName "\"$Name $Supplement\""
    } else {
      set AttrName "\"$Name\""
    }

    # Set attribute type to floating-point
    set type 2

    # Append floating-point attributes
    if {$i > 0} {append text "\n"}
    append text "\n[format $fmt0 $AttrName]"
    append text "\n[format $fmt1 $type]"
    append text "\n[format $fmt2 "\{ \[${i}.0\] \}"]"

    # Increment attribute counter
    incr i
  }

  # Open temporary parameter file
  if [catch {set fd [open $TmpParFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A  [::msgcat::mc $fmt $TmpParFile]
    return 0
  }

  # Write temporary parameter file
  chan puts $fd $text

  # Close temporary parameter file
  catch {chan close $fd}

  # Check size of the temporary parameter file
  if {[file exists $TmpParFile] && [file size $TmpParFile] == 0} {
    set fmt "Parameter file %s incomplete"
    monitor_msg $w A [::msgcat::mc $fmt $TmpParFile]
    return 0
  }

  # Create log file
  set LogFile [TmpFileName log txt tmp]

  # Check log file
  if {$LogFile eq ""} {
    set msg "Log file cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Open log file
  if [catch {set fd_log [open $LogFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt $LogFile]
    return 0
  }

  # Add the information in the day file to the operational data file(s)
  set fd_backend [open "| calculo $TmpParFile |& cat" "r+"]
  chan configure $fd_backend -buffering line

  # Set handler for a channel event
  chan event $fd_backend readable "fhandler $fd_backend $fd_log"

  # Stay in an event loop until the update process finishes
  vwait ReturnCode

  # Check log file
  if {[file exists $LogFile] && [file size $LogFile] == 0} {
    file delete $LogFile
    set msg "Log file could not be written"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Print error message, if necessary
  if {$ReturnCode != 0} {
    # Show error message
    set msg "Error during set-up of the estimated submodel"
    monitor_msg $w A [::msgcat::mc $msg]
    update
 
    # Show log file
    set Title [::msgcat::mc "Set-up of Estimated Submodel"]
    set TagList {{text "-justify left"}}
    set TextWin [TextViewer_create .fv 82 40 $Title $TagList]
    text_loadFile $TextWin $LogFile text
    tkwait window .fv
    catch {file delete $LogFile}
    exit
  }

  # Delete log file
  catch {file delete $LogFile}

  # Display concluding message
  monitor_msg $w F $TaskName

  return 1
}

#-----------------------------------------------------------------------

proc buildShadowGradeOctree {SubMod ParArray w} {

#***********************************************************************
#
# Function
#
#     Front end for merging an estimated submodel with an existing
#     grade octree.
#
#     The procedure runs program mergo as a backend process.
#
# Arguments
#
#     SubMod   : estimated submodel
#     ParArray : array of update parameters
#     w        : name of a monitor window
#
# Global Variable
#
#     ReturnCode
#
# Return Value
#
#      1 : success
#      0 : failure
#
# Reqd. Procedures
#
#     fhandler, monitor_msg, text_loadFile, TextViewer_create,
#     TmpFileName
#
# Date written  : 14/05/2000 by R. Prissang (FUB)
# Last modified : 13/03/2014 by R. Prissang (FUB)
#
#***********************************************************************

  upvar $ParArray Par

  global ReturnCode

  # Check submodel
  if {[string trim $SubMod] eq ""} {
    error "No estimated submodel given"
  }

  # Check parameter array
  if {! [array exists Par]} {
    error "Variable Par is no array or it does not exist"
  }

  # Check monitor widget
  if {[string trim $w] eq ""} {
    error "No monitor window pathname given"
  }

  # Initialise variables
  set ReturnCode ""
  set OutFile "$Par(GradeDir)/$Par(ShadowOctree)"
  set GradeOctree "$Par(GradeDir)/$Par(GradeOctree)" 
  set Params "\"$SubMod\" \"$GradeOctree\" \"$OutFile\""

  # Display start message
  set Taskname [::msgcat::mc "Building new grade octree"]
  monitor_msg $w S $Taskname

  # Create log file
  set LogFile [TmpFileName log txt tmp]

  # Check log file
  if {$LogFile eq ""} {
    set msg "Log file cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return 0
  }

  # Open log file
  if [catch {set fd_log [open $LogFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt $LogFile]
    return 0
  }

  # Open pipeline and start the merging program
  set fd_backend [open "| mergo $Params |& cat" "r+"]
  chan configure $fd_backend -buffering line

  # Set handler for a channel event
  chan event $fd_backend readable "fhandler $fd_backend $fd_log"

  # Stay in an event loop until the update process finishes
  vwait ReturnCode

  # Check log file (no output by mergo except error messages)
  if {$ReturnCode != 0} {
    if {[file exists $LogFile] && [file size $LogFile] == 0} {
      file delete $LogFile
      set msg "Log file could not be written"
      monitor_msg $w A [::msgcat::mc $msg]
      return 0
    }
  }

  # Print error message, if necessary
  if {$ReturnCode != 0} {
    # Show error message
    set msg "Error during building a new grade octree"
    monitor_msg $w A [::msgcat::mc $msg]
    update

    # Show log file
    set Title [::msgcat::mc "Build Shadow Grade Octree"]
    set TagList {{text "-justify left"}}
    set TextWin [TextViewer_create .fv 82 40 $Title $TagList]
    text_loadFile $TextWin $LogFile text
    tkwait window .fv
    catch {file delete $LogFile}
    exit
  }

  # Delete log file
  catch {file delete $LogFile}

  # Display concluding message
  monitor_msg $w F $Taskname

  return 1
}

#-----------------------------------------------------------------------

proc estoFrontEnd {w ParFile TaskName GaugeWin} {

#***********************************************************************
#
# Function
#
#     Front end for program esto. The procedure runs esto as a backend
#     process. Front handlers creating and updating a progress gauge
#     are called whenever esto writes to stdout.
#
# Arguments
#
#     w        : name of a monitor window
#     ParFile  : parameter file
#     TaskName : task description
#     GaugeWin : name of a gauge window
#
# Global Variable
#
#     ReturnCode
#
# Return Value
#
#      1 : success
#      0 : error
#     -1 : abort
#
# Reqd. Procedures
#
#     gfhandler, monitor_msg, text_loadFile, text_addLineTags,
#     TextViewer_create, TmpFileName
#
# Date written  : 14/05/2000 by R. Prissang (FUB)
# Last modified : 13/03/2014 by R. Prissang (FUB)
#
#***********************************************************************

  global ReturnCode

  # Initialise variables
  set ReturnCode ""

  # Display start message
  set fmt "Estimation %s"
  monitor_msg $w S [::msgcat::mc $fmt $TaskName]

  # Create log file
  set LogFile [TmpFileName log txt tmp]

  # Check log file
  if {$LogFile eq ""} {
    set msg "Log file cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return -1
  }

  # Open log file
  if [catch {set fd_log [open $LogFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt $LogFile]
    return -1
  }

  # Open pipeline and start estimation program (OK)
  set backend [open "| esto $ParFile |& cat" "r+"]
  chan configure $backend -buffersize 1 -buffering none

  # Set event handler
  chan event $backend readable "gfhandler $backend $GaugeWin $fd_log"

  # Stay in an event loop until the estimation process finishes
  vwait ReturnCode

  # Check log file
  if {[file exists $LogFile] && [file size $LogFile] == 0} {
    file delete $LogFile
    set msg "Log file could not be written"
    monitor_msg $w A [::msgcat::mc $msg]
    return -1
  }

  # Print error message, if necessary
  if {$ReturnCode != 0} {
    # Show error message
    destroy .gt
    set msg "Error during property estimation"
    monitor_msg $w A [::msgcat::mc $msg]
    update

    # Show log file
    set TagList {{text "-justify left"} {red "-foreground red"}}
    set TextWin [TextViewer_create .fv 82 40 $TaskName $TagList]
    text_loadFile $TextWin $LogFile text
    text_addLineTags $TextWin "ot correctly" red
    text_addLineTags $TextWin "ot able" red
    text_addLineTags $TextWin "atal error" red
    text_addLineTags $TextWin "used=0" red
    tkwait window .fv
    catch {file delete $LogFile}
    exit
  }

  # Delete log file
  catch {file delete $LogFile}

  set fmt "Estimation %s"
  monitor_msg $w F [::msgcat::mc $fmt $TaskName]
  return 1
}

#-----------------------------------------------------------------------

proc extractUnknown {ParFile w} {

#***********************************************************************
#
# Function
#
#     Front end for the extraction of obels with an unknown attribute
#     from an octree. The procedure runs program edito as a backend
#     process.
#
# Arguments
#
#     ParFile  : parameter file
#     w        : name of a monitor window
#
# Global Variable
#
#     ReturnCode
#
# Return Value
#
#      1 : success
#      0 : error
#     -1 : abort
#
# Reqd. Procedures
#
#     fhandler, monitor_msg, text_loadFile, TextViewer_create,
#     TmpFileName
#
# Date written  : 14/05/2000 by R. Prissang (FUB)
# Last modified : 18/01/2015 by R. Prissang (FUB)
#
#***********************************************************************

  global ReturnCode

  # Initialise variables
  set ReturnCode ""
  set TaskName [::msgcat::mc "Extraction (unknown blocks)"]

  # Display start message
  monitor_msg $w S $TaskName

  # Check, if the parameter file for the region extraction exists
  if {! [file exists $ParFile]} {
    set msg "Parameter file %s not found"
    monitor_msg $w A [::msgcat::mc $msg $ParFile]
    return -1
  }

  # Create log file
  set LogFile [TmpFileName log txt tmp]

  # Check log file
  if {$LogFile eq ""} {
    set msg "Log file cannot be created"
    monitor_msg $w A [::msgcat::mc $msg]
    return -1
  }

  # Open log file
  if [catch {set fd_log [open $LogFile w]}] {
    set fmt "Can't open file %s"
    monitor_msg $w A [::msgcat::mc $fmt $LogFile]
    return -1
  }

  # Add the information in the day file to the operational data file(s)
  set fd_backend [open "| edito $ParFile |& cat" "r+"]
  chan configure $fd_backend -buffering line

  # Set handler for a channel event
  chan event $fd_backend readable "fhandler $fd_backend $fd_log"

  # Stay in an event loop until the update process finishes
  vwait ReturnCode

  # Check log file
  if {[file exists $LogFile] && [file size $LogFile] == 0} {
    file delete $LogFile
    set msg "Log file could not be written"
    monitor_msg $w A [::msgcat::mc $msg]
    return -1
  }

  # Print error message, if necessary
  if {$ReturnCode != 0} {
    # Show error message
    set msg "Error during the extraction of unknown blocks"
    monitor_msg $w A [::msgcat::mc $msg]
    update

    # Show log file
    set Title [::msgcat::mc "Extraction of Unknown Blocks"]
    set TagList {{text "-justify left"}}
    set TextWin [TextViewer_create .fv 82 40 $Title $TagList]
    text_loadFile $TextWin $LogFile text
    tkwait window .fv
    catch {file delete $LogFile}
    return 0
  }

  # Open log file to extract the number ob obels
  if {! [catch {set fd_log [open $LogFile r]}]} {
    # Check if any unknown obels have been extracted
    while {[chan gets $fd_log line] >= 0} {
      # Find line stating the number of obels written
      if [regexp {^Obels written *: *[0-9]? *} $line] {
        # Read number of obels
        scan [regsub {Obels written *: *} $line {}] "%s" nObels

        # Build concluding message
        switch $nObels {
          0 { # No unknown blocks left
            set msg [::msgcat::mc "Extraction (no unknown block left)"]
          }
          1 { # One unknown block extracted
            set msg [::msgcat::mc "Extraction (one unknown block)"]
          }
          default { # More than one blocks extracted
            set fmt "Extraction (%s unknown blocks)"
            set msg [::msgcat::mc $fmt [LocalNum $nObels]]
          }
        }
        # Update task name
        set TaskName $msg
      }
    }

    # Close log file
    catch {chan close $fd_log}
  }

  # Delete log file
  catch {file delete $LogFile}

  # Display concluding message
  monitor_msg $w F $TaskName

  return 1
}
